from typing import Union

import torch as th
from torch import Tensor

from .gem_torch import gem_torch

__all__ = ["gem"]


_triton_available = True


def gem(x: Tensor, p: Union[float, Tensor] = 3.0, eps: float = 1e-6, dim: int = -2, keepdim=True):
    global _triton_available

    if x.device == th.device("cpu"):
        return gem_torch(x, p, eps, dim, keepdim)

    try:
        assert _triton_available
        from .gem_triton import GeMOps

        if not x.is_contiguous():
            x = x.contiguous()
        if not p.is_contiguous():
            p = p.contiguous()

        return GeMOps.apply(x, p, eps, dim, keepdim)
    except:
        _triton_available = False
        return gem_torch(x, p, eps, dim, keepdim)



from typing import Union

import torch as th
import torch.nn as nn
from torch import Tensor

from .functional import gem


class GeM(nn.Module):
    def __init__(
        self,
        p: Union[float, int, Tensor] = 3.0,
        eps: float = 1e-6,
        p_trainable: bool = True,
        dim: int = -2,
        keepdim=True,
    ):
        """
        Generalized Mean (GeM) Pooling Layer.

        This layer applies the Generalized Mean (GeM) pooling operation over an input tensor.
        GeM is a generalization of average and max pooling, controlled by the parameter `p`.

        Args:
        - p (float | int | Tensor): The pooling parameter. It can be trainable by using `p_trainable=True` (default=True).
        - eps (float): A small value added to the input tensor to avoid numerical instability.
        - p_trainable (bool): If True, `p` is a learnable parameter. Otherwise, `p` is fixed.
        - dim (int): The dimension over which the pooling operation is applied.
            Default is -2, this will reduce last 2 dimensions so usually useful with image tensor.
            For 1D tensor like audio and 4D tensor like voxel, you can use dim=-1 and dim=-3 respectively.
        """
        super().__init__()

        if p_trainable:
            if isinstance(p, Tensor):
                self.p = nn.Parameter(p.to(th.float))
            elif isinstance(p, (int, float)):
                self.p = nn.Parameter(th.tensor(p, dtype=th.float))
            else:
                raise NotImplementedError(f"Unknown data type for `p`: {type(p)}")
        else:
            self.p = p

        self.eps = eps
        self.dim = dim
        self.keepdim = keepdim

    def forward(self, x: Tensor, dim=-2):
        return gem(x, self.p, self.eps, dim=dim, keepdim=self.keepdim)

    def __repr__(self):
        p_val = self.p
        if isinstance(p_val, Tensor):
            p_val = p_val.item()

        return f"{self.__class__.__name__}(p={p_val:.4f}), eps={self.eps})"



from typing import Union

import torch as th
import triton
import triton.language as tl
from torch import Tensor

from fast_gem.functional import triton_utils as tu


class GeMOps(th.autograd.Function):
    @staticmethod
    @triton.jit
    def forward_kernel(x_ptr, y_ptr, p, eps, M, N, IS_P_TENSOR: tl.constexpr, BLK_M: tl.constexpr, BLK_N: tl.constexpr):
        pid = tl.program_id(0)
        offs_m = pid * BLK_M + tl.arange(0, BLK_M)  # m
        offs_n = tl.arange(0, BLK_N)  # n
        x_ptrs = x_ptr + offs_m[:, None] * N + offs_n[None, :]  # m n

        if IS_P_TENSOR:
            p = tl.load(p)

        y = tl.zeros((BLK_M,), dtype=tl.float32)
        for n in range(tl.cdiv(N, BLK_N)):
            mask = (offs_m[:, None] < M) & (offs_n[None, :] < N - n * BLK_N)
            x = tl.load(x_ptrs, mask=mask, other=0.0)  # m n

            # calculate adaptive average pooling
            x = tl.where((x < eps) & mask, eps, x)
            x = tu.pow(x, p)  # m n
            y += tl.sum(x, 1)  # m

            x_ptrs += BLK_N

        y /= N
        y = tu.pow(y, 1 / p)

        y_ptrs = y_ptr + offs_m
        tl.store(y_ptrs, y, mask=offs_m < M)

    @staticmethod
    @triton.jit
    def backward_kernel(
        x_ptr, y_ptr, p, dy_ptr, dx_ptr, dp_ptr, eps, M, N, IS_P_TENSOR: tl.constexpr, BLK_M: tl.constexpr, BLK_N: tl.constexpr
    ):
        pid = tl.program_id(0)
        offs_m = pid * BLK_M + tl.arange(0, BLK_M)  # m
        offs_n = tl.arange(0, BLK_N)  # n
        x_ptrs = x_ptr + offs_m[:, None] * N + offs_n[None, :]  # m n
        dx_ptrs = dx_ptr + offs_m[:, None] * N + offs_n[None, :]  # m n
        mask_m = offs_m < M

        if IS_P_TENSOR:
            p = tl.load(p)

        # calculate y-level grad
        y = tl.load(y_ptr + offs_m, mask=mask_m)
        dy = tl.load(dy_ptr + offs_m, mask=mask_m)

        if IS_P_TENSOR:
            dp = tl.where(mask_m, -tl.log(y) / p * y * dy, 0.0)
            dp = tl.sum(dp, axis=0)
        dy = dy / p * y / (tu.pow(y, p) * N)  # m

        # re-calculate x for x-level grad
        for n in range(tl.cdiv(N, BLK_N)):
            mask = mask_m[:, None] & (offs_n[None, :] < N - n * BLK_N)
            x = tl.load(x_ptrs, mask=mask, other=0.0)  # m n

            # calculate adaptive average pooling
            x_ = tl.where((x < eps) & mask, eps, x)
            # x = tu.pow(x, p)
            # y += tl.sum(x, 1)  # m

            dx = tl.zeros((BLK_M, BLK_N), dtype=dy.dtype) + dy[:, None]  # m n

            if IS_P_TENSOR:
                dp_tmp = tl.where(mask, tu.pow(x_, p) * tl.log(x_) * dx, 0.0)
                dp += tl.sum(tl.sum(dp_tmp, axis=1), axis=0)

            dx *= p * tu.pow(x_, p - 1)
            dx = tl.where((x < eps) & mask, 0.0, dx)

            tl.store(dx_ptrs, dx, mask=mask)
            x_ptrs += BLK_N
            dx_ptrs += BLK_N

        if IS_P_TENSOR:
            tl.atomic_add(dp_ptr, dp)

    @staticmethod
    def forward(ctx, x: Tensor, p: Union[float, Tensor] = 3.0, eps: float = 1e-6, dim: int = -2, keepdim=True):
        ctx.is_p_tensor = isinstance(p, Tensor)
        dim = x.ndim + dim if dim < 0 else dim
        M = tu.cummul(*x.shape[:dim])
        N = tu.cummul(*x.shape[dim:])
        y = x.new_empty(list(x.shape[:dim]) + ([1 for _ in range(x.ndim - dim)] if keepdim else []))

        BLK_N = max(min(triton.next_power_of_2(N), 4096), 32)
        BLK_M = max(1, 1024 // BLK_N)
        grid = lambda meta: (triton.cdiv(M, meta["BLK_M"]),)
        GeMOps.forward_kernel[grid](x, y, p, eps, M, N, IS_P_TENSOR=ctx.is_p_tensor, BLK_M=BLK_M, BLK_N=BLK_N)

        if ctx.is_p_tensor:
            ctx.save_for_backward(x, p, y)
            ctx.params = eps, dim
        else:
            ctx.save_for_backward(x, y)
            ctx.params = p, eps, dim
        return y

    @staticmethod
    def backward(ctx, dy: Tensor):
        if ctx.is_p_tensor:
            x, p, y = ctx.saved_tensors
            eps, dim = ctx.params
        else:
            x, y = ctx.saved_tensors
            p, eps, dim = ctx.params

        M = tu.cummul(*x.shape[:dim])
        N = tu.cummul(*x.shape[dim:])
        dx = th.empty_like(x)
        dp = None
        if ctx.is_p_tensor:
            dp = th.zeros_like(p)

        BLK_N = max(min(triton.next_power_of_2(N), 4096), 32)
        BLK_M = max(1, 1024 // BLK_N)
        grid = lambda meta: (triton.cdiv(M, meta["BLK_M"]),)
        GeMOps.backward_kernel[grid](x, y, p, dy, dx, dp, eps, M, N, IS_P_TENSOR=ctx.is_p_tensor, BLK_N=BLK_N, BLK_M=BLK_M)
        return dx, dp, None, None, None



from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as f:
    desc = f.read()

setup(
    name="fast_gem",
    version="0.0.4",
    description="Efficient and general implementation of Generalized Mean Pooling (GeM)",
    author="Kitsunetic",
    author_email="jh.shim.gg@gmail.com",
    url="https://github.com/Kitsunetic/fast-GeM",
    packages=find_packages(),
    zip_safe=False,
    install_requires=[],
    # entry_points={"console_scripts": ["kitsu=kitsu.main:main"]},
    long_description=desc,
    long_description_content_type="text/markdown",
)



import torch as th
import torch.nn.functional as F
from kitsu.utils import cummul
from torch import Tensor


def gem_torch_old(x, p=3, eps=1e-6, work_with_tokens=False):
    if work_with_tokens:
        x = x.permute(0, 2, 1)
        # unseqeeze to maintain compatibility with Flatten
        return F.avg_pool1d(x.clamp(min=eps).pow(p), (x.size(-1))).pow(1.0 / p).unsqueeze(3)
    else:
        return F.avg_pool2d(x.clamp(min=eps).pow(p), (x.size(-2), x.size(-1))).pow(1.0 / p)


def gem_torch(x: Tensor, p=3, eps=1e-6, dim=-2, keepdim=True):
    """
    input:
    - x: ... ... (= m n)
    return:
    - y: ... (= m)
    """
    dim = x.ndim + dim if dim < 0 else dim
    yshape = list(x.shape[:dim]) + ([1 for _ in range(x.ndim - dim)] if keepdim else [])

    M = cummul(*x.shape[:dim])
    N = cummul(*x.shape[dim:])
    x = x.view(M, N)  # m n

    x = x.clamp(eps).pow_(p)
    x = x.mean(-1)  # adaptive average pooling
    x = x.pow_(1.0 / p)

    return x.view(*yshape)
